from django.db import models
from django.dispatch import Signal

post_update = Signal()
post_bulk_create = Signal()


class CustomQuerySet(models.query.QuerySet):
    def update(self, **kwargs):
        # fetching objects to be updated before update happens, since update might affect the queryset
        objs = list(self)
        super(CustomQuerySet, self).update(**kwargs)
        kwargs.update({'objs': objs})
        post_update.send(sender=self.model, **kwargs)


class CustomManager(models.Manager):
    def get_queryset(self):
        return CustomQuerySet(self.model, using=self._db)

    def bulk_create(self, objs, batch_size=None):
        super(CustomManager, self).bulk_create(objs, batch_size)
        kwargs = {'objs': objs, 'batch_size': batch_size}
        post_bulk_create.send(sender=self.model, **kwargs)



from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.models import Permission
from django.shortcuts import resolve_url
from .utils import current_user
from django.core.exceptions import PermissionDenied
from functools import wraps
from urllib.parse import urlparse

def get_user_permissions(user):
    return user.user_permissions.all() | Permission.objects.filter(group__user=user)


def get_permissions_strings(user):
    return list({"%s.%s" % (ct, name) for ct, name in
                 get_user_permissions(user).values_list('content_type__app_label', 'codename')})


def get_current_user_permissions():
    return get_user_permissions(current_user())


def get_current_user_permissions_strings():
    return get_permissions_strings(current_user())


def user_has_perms(user, perms, exclude_superuser=False):
    """
    Checks if the user has at least one perm in the list of perms
    :param user:
    :param perms: Perms to check
    :param exclude_superuser: By default, superuser has all perms. If this is True, then it checks permissions list.
    :return: True if user has at least one perm
    """
    if not exclude_superuser and user.is_superuser:
        return True
    if exclude_superuser:
        user_permissions = get_permissions_strings(user)
        has_perms = [perm in user_permissions for perm in perms]
    else:
        has_perms = [user.has_perm(perm) for perm in perms]
    return not not list(filter(None, has_perms))


def current_user_has_perms(perms, exclude_superuser=False):
    return user_has_perms(current_user(), perms, exclude_superuser)


def get_users_for_perms(perms):
    users = set()
    for perm in perms:
        app_label, perm_name = perm.split('.')
        permission = Permission.objects.filter(content_type__app_label=app_label, codename=perm_name).first()
        if permission:
            users.update(list(permission.user_set.all()))
            for group in permission.group_set.all():
                users.update(list(group.user_set.all()))
    return list(users)


def is_superuser(user):
    return user.is_active and user.is_staff and user.is_superuser


def user_passes_test(test_func, login_url=None, redirect_field_name=REDIRECT_FIELD_NAME, raise_exception=False):
    """
    Decorator for views that checks that the user passes the given test,
    redirecting to the log-in page if necessary. The test should be a callable
    that takes the user object and returns True if the user passes.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            if test_func(request.user):
                return view_func(request, *args, **kwargs)
            path = request.build_absolute_uri()
            resolved_login_url = resolve_url(login_url or settings.LOGIN_URL)
            # If the login url is the same scheme and net location then just
            # use the path as the "next" url.
            login_scheme, login_netloc = urlparse(resolved_login_url)[:2]
            current_scheme, current_netloc = urlparse(path)[:2]
            if ((not login_scheme or login_scheme == current_scheme) and
                    (not login_netloc or login_netloc == current_netloc)):
                path = request.get_full_path()
            from django.contrib.auth.views import redirect_to_login
            if raise_exception:
                raise PermissionDenied
            return redirect_to_login(
                path, resolved_login_url, redirect_field_name)

        return _wrapped_view

    return decorator



def suit_multi_admin():
    """Monkeypatch Django Suit to support multiple Admin Sites

    Usage:

        put this code for example in myapp/monkey.py

        Activate the monkeypatch in your models.py:

        from monkey import suit_multi_admin
        suit_multi_admin()

        Put a separate configuration in your settings as:
        SUIT_CONFIG_<MYADMIN> = {...}
        where MYADMIN is the upper case name of your custom AdminSite.name

    """

    import suit.config
    from suit.templatetags import suit_menu
    from django.conf import settings
    from django.core.handlers.wsgi import WSGIRequest

    def get_config(param=None, config_key='SUIT_CONFIG'):
        """
        Accept a custom config_key kwarg
        """
        if hasattr(settings, config_key):
            config = getattr(settings, config_key, {})
        else:
            config = suit.config.default_config()
        if param:
            value = config.get(param)
            if value is None:
                value = suit.config.default_config().get(param)
            return value
        return config

    suit.config.get_config = get_config

    @suit_menu.register.simple_tag(takes_context=True)
    def get_menu(context, request):
        """
        Pass admin_site_name go Menu via context
        """
        if not isinstance(request, WSGIRequest) or request.user.is_anonymous:
            return None

        # Try to get app list
        if hasattr(request, 'current_app'):
            # Django 1.8 uses request.current_app instead of context.current_app
            admin_site = suit_menu.get_admin_site(request.current_app)
        else:
            try:
                admin_site = suit_menu.get_admin_site(context.current_app)
            # Django 1.10 removed the current_app parameter for some classes and functions.
            # Check the release notes.
            except AttributeError:
                admin_site = suit_menu.get_admin_site(context.request.resolver_match.namespace)

        # Try to get app list
        template_response = admin_site.index(request)
        context['admin_site_name'] = admin_site.name
        try:
            app_list = template_response.context_data['app_list']
        except Exception:
            return
        menu = suit_menu.Menu(context, request, app_list)
        menu_app_list = [app for app in menu.get_app_list() if app.get('models') or app.get('url')]
        if menu.conf_menu_extras:
            menu_app_list.extend(menu.make_menu(menu.conf_menu_extras))
        if getattr(settings, 'REUSABLE_SHOW_ERRORS_LINK', False):
            menu_app_list.extend(menu.make_menu((
                {
                    'label': 'Errors',
                    'url': 'django_reusable:errors',
                    'permissions': getattr(settings, 'REUSABLE_SHOW_ERRORS_PERM', 'django_reusable.view_errors')
                },
            )))
        return menu_app_list

    suit_menu.get_menu = get_menu

    def patched_init(self, context, request, app_list):
        """
        Get admin_site_name from context
        """
        self.request = request
        self.app_list = app_list
        self.admin_site_name = context.get('admin_site_name')

        # Detect current app, if any
        try:
            self.ctx_app = context['app_label'].lower()
        except Exception:
            self.ctx_app = None

        # Get current model plural name, if any
        try:
            self.ctx_model_plural = context['opts'].verbose_name_plural.lower()
        except Exception:
            self.ctx_model_plural = None

        # Flatten all models from native apps
        self.all_models = [model for app in app_list for model in app['models']]

        # Init config variables
        self.init_config()
        super(suit_menu.Menu, self).__init__()

    suit_menu.Menu.__init__ = patched_init

    def patched_init_config(self):
        """
        Set config_key based on admin_site_name.
        """
        if self.admin_site_name and self.admin_site_name != 'admin':
            config_key = 'SUIT_CONFIG_{}'.format(self.admin_site_name.upper())
        else:
            config_key = 'SUIT_CONFIG'
        self.conf_exclude = get_config('MENU_EXCLUDE', config_key)
        self.conf_open_first_child = get_config('MENU_OPEN_FIRST_CHILD', config_key)
        self.conf_icons = get_config('MENU_ICONS', config_key)
        self.conf_menu_order = get_config('MENU_ORDER', config_key)
        self.conf_menu = get_config('MENU', config_key)
        self.conf_menu_extras = get_config('MENU_EXTRAS', config_key)

    suit_menu.Menu.init_config = patched_init_config

    original_app_is_forbidden = suit_menu.Menu.app_is_forbidden

    def patched_is_forbidden(self, entity):
        if entity['permissions'] and callable(entity['permissions']):
            return not entity['permissions'](self.request.user)
        return original_app_is_forbidden(self, entity)

    suit_menu.Menu.app_is_forbidden = patched_is_forbidden
    suit_menu.Menu.model_is_forbidden = patched_is_forbidden



from django_tables2 import Table
from django_tables2.rows import BoundRow, BoundRows


class EnhancedBoundRow(BoundRow):

    def new_rows(self):
        new_row_columns = self.table.new_row_columns
        rows = sorted(set(new_row_index for column, new_row_index in new_row_columns))
        for row in rows:
            yield self.get_new_row_items(row, new_row_columns)

    def get_new_row_items(self, row, new_row_columns):
        for column, new_row_index in new_row_columns:
            if row == new_row_index:
                column.current_value = self.get_cell(column.name)
                column.current_record = self.record
                if not column.current_value and getattr(column.column, 'no_empty_cell', False):
                    continue
                yield column, column.current_value


class EnhancedBoundRows(BoundRows):
    def __iter__(self):
        # Top pinned rows
        for pinned_record in self.generator_pinned_row(self.pinned_data.get('top')):
            yield pinned_record

        for record in self.data:
            yield EnhancedBoundRow(record, table=self.table)

        # Bottom pinned rows
        for pinned_record in self.generator_pinned_row(self.pinned_data.get('bottom')):
            yield pinned_record

    def __getitem__(self, key):
        """
        Slicing returns a new `~.BoundRows` instance, indexing returns a single
        `~.BoundRow` instance.
        """
        if isinstance(key, slice):
            return self.__class__(
                data=self.data[key],
                table=self.table,
                pinned_data=self.pinned_data
            )
        else:
            return EnhancedBoundRow(record=self.data[key], table=self.table)


class EnhancedTable(Table):

    def __init__(self, *args, extra_data={}, fields=None, **kwargs):
        kwargs['template_name'] = 'django_reusable/tables/enhanced-table.html'
        self.extra_data = extra_data
        super().__init__(*args, **kwargs)
        if self.new_row_columns:
            self.rows = EnhancedBoundRows(data=self.data, table=self, pinned_data=self.pinned_data)
        if fields is not None:
            all_fields = self.base_columns.keys()
            self.exclude = self.exclude or []
            self.exclude.extend(set(all_fields).difference(fields))
        self.extra_footers = self.get_extra_footers()

    @property
    def new_row_columns(self):
        return [(column, column.column.new_row_index) for column in self.columns.all()
                if getattr(column.column, 'has_new_row_index', False)]

    def get_extra_footers(self):
        """
        :return: List of dict with column as key for each footer row
        """
        return []



import json
import traceback
from ast import literal_eval
from collections import namedtuple

from django.contrib.contenttypes.models import ContentType
from django.core.paginator import Paginator, EmptyPage
from django.db import models
from django.urls import NoReverseMatch, reverse
from django.utils.safestring import mark_safe
from django.utils.timezone import now

from django_reusable.admin.urls import ModelURLs
from django_reusable.utils import get_property


class GenericRelationMixin(models.Model):
    @property
    def content_object_link(self):
        return mark_safe(f'''{self.content_type}: <a target="_blank" 
        href="{self.content_object.get_obj_change_url()}">{self.content_object}</a>''')

    @property
    def this_content_object(self):
        return type(self).objects.filter(content_type=self.content_type, object_id=self.object_id)

    @property
    def this_content_type(self):
        return type(self).objects.filter(content_type=self.content_type)

    @classmethod
    def for_object(cls, obj):
        return cls.objects.filter(content_type=ContentType.objects.get_for_model(type(obj)), object_id=obj.id)

    @classmethod
    def for_objects(cls, objs):
        return cls.objects.filter(content_type=ContentType.objects.get_for_model(type(objs[0])),
                                  object_id__in=[obj.id for obj in objs]) if objs else None

    @classmethod
    def for_object_type(cls, obj):
        return cls.objects.filter(content_type=ContentType.objects.get_for_model(type(obj)))

    @classmethod
    def for_model(cls, model):
        return cls.objects.filter(content_type=ContentType.objects.get_for_model(model))

    @staticmethod
    def model_content_type(model):
        return ContentType.objects.get_for_model(model)

    @staticmethod
    def obj_content_type(obj):
        return ContentType.objects.get_for_model(type(obj))

    class Meta:
        abstract = True


class ModelUtilsMixin:

    @classmethod
    def get_add_url_name(cls, site='admin'):
        return ModelURLs(cls, site).get_add_url_name()

    @classmethod
    def get_add_url(cls, site='admin'):
        return ModelURLs(cls, site).get_add_url()

    @classmethod
    def get_changelist_url_name(cls, site='admin'):
        return ModelURLs(cls, site).get_changelist_url_name()

    @classmethod
    def get_changelist_url(cls, site='admin'):
        return ModelURLs(cls, site).get_changelist_url()

    @classmethod
    def get_change_url_name(cls, site='admin'):
        return ModelURLs(cls, site).get_change_url_name()

    def get_change_url_for_object(self, site='admin'):
        return ModelURLs(self, site, self.pk).get_obj_change_url()

    @classmethod
    def get_change_url(cls, pk, site='admin'):
        return ModelURLs(cls, site, pk).get_obj_change_url()

    def get_obj_change_link_tag(self, text, css_class='', target='', site='admin'):
        return ModelURLs(self, site, self.pk).get_obj_change_link_tag(text, css_class, target)

    def get_obj_change_url(self, site='admin'):
        return ModelURLs(self, site, self.pk).get_obj_change_url()

    @classmethod
    def get_delete_url_name(cls, site='admin'):
        return ModelURLs(cls, site).get_delete_url_name()

    @classmethod
    def get_delete_url(cls, pk, site='admin'):
        return ModelURLs(cls, site, pk).get_obj_delete_url()

    def get_obj_delete_url(self, site='admin'):
        return ModelURLs(self, site, self.pk).get_obj_delete_url()

    @property
    def change_link(self):
        return ModelURLs(self, pk=self.pk).get_change_link()

    @property
    def original_obj(self):
        if not hasattr(self, '_orig_obj'):
            self._orig_obj = self.__class__.objects.get(id=self.pk) if self.pk else None
        return self._orig_obj

    def get_property(self, name, fn):
        return get_property(self, name, fn)

    class Meta:
        abstract = True


class TimeStampedModelOnly(models.Model, ModelUtilsMixin):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class MonitorChangeMixin:
    """
        List of tuples:
        ('field_name', dict(callback: lambda (self, old_value, new_value)))
    """
    monitor_fields = []

    def notify_changes(self):
        try:
            for field, options in self.monitor_fields:
                if not options:
                    continue
                old_value = getattr(self.original_obj, field) if self.original_obj else None
                new_value = getattr(self, field)
                if old_value != new_value:
                    callback = options.get('callback', None)
                    if callback:
                        callback(self, old_value, new_value)
        except:
            traceback.print_exc()

    class Meta:
        abstract = True


class TimeStampedModel(TimeStampedModelOnly, MonitorChangeMixin):
    class Meta:
        abstract = True


Page = namedtuple("Page", "has_next, next_num, has_prev, prev_num, items ")


class Error(models.Model):
    hash = models.CharField(max_length=64, primary_key=True)
    host = models.CharField(max_length=1024)
    path = models.CharField(max_length=4096)
    method = models.CharField(max_length=64)
    request_data = models.TextField()
    exception_name = models.CharField(max_length=256)
    traceback = models.TextField()
    count = models.IntegerField(default=0)
    created_on = models.DateTimeField(auto_now=True)
    last_seen = models.DateTimeField(auto_now=True, db_index=True)

    @property
    def request_data_json(self):
        return json.dumps(literal_eval(self.request_data), indent=4)

    @classmethod
    def get_exceptions_per_page(cls, page_number=1):
        records = cls.objects.all()
        paginator = Paginator(records, 25)
        try:
            page = paginator.page(page_number)
            return Page(page.has_next(),
                        page.next_page_number() if page.has_next() else None,
                        page.has_previous(),
                        page.previous_page_number() if page.has_previous() else None,
                        page.object_list)
        except EmptyPage:
            return Page(False, None, True, paginator.num_pages, [])

    @classmethod
    def get_entity(cls, rhash):
        return cls.objects.get(pk=rhash)

    @classmethod
    def create_or_update_entity(cls, rhash, host, path, method, request_data, exception_name, _traceback):
        try:
            obj, created = cls.objects.get_or_create(hash=rhash)
            if created:
                obj.host, obj.path, obj.method, obj.request_data, obj.exception_name, obj.traceback = \
                    host, path, method, request_data, exception_name, _traceback
                obj.count = 1
                obj.save()
            else:
                obj.count += 1
                obj.last_seen = now()
                obj.save(update_fields=['count', 'last_seen'])
        except:
            traceback.print_exc()

    @classmethod
    def delete_entity(cls, rhash):
        return cls.objects.filter(pk=rhash).delete()

    def __str__(self):
        return "'%s' '%s' %s" % (self.host, self.path, self.count)

    def __unicode__(self):
        return self.__str__()

    class Meta:
        ordering = ['-last_seen']



import inspect
import re

from django.conf import settings

from django_reusable.logging.loggers import PrintLogger
from django_reusable.utils import spaces
from django_reusable.utils.file_utils import generate_file_if_updated

try:
    from dajaxice.core import dajaxice_functions
except ImportError:
    dajaxice_functions = None

AUTO_GENERATED_COMMENT = '// This file is auto generated. Please do not edit. Your changes will be overwritten.\n'

DAJAXICE_STATIC_TYPES = '''
import { DajaxiceFn } from "dry-ux";

export const modules: DajaxiceModules = {};\n\n'''


def generate_app_path_enums():
    file_path = getattr(settings, 'REUSABLE_APP_URL_TS_INTERFACE_PATH', None)
    if not file_path:
        return
    logger = PrintLogger("[django_reusable] generate_app_path_enums")

    def replace_variable(url):
        new_url = re.sub(r"%\((\w+)\)s", r":\1", url)
        return new_url

    from django_reusable.urls.utils import get_all_urls
    app_urls = ',\n'.join([f'{spaces(4)}"{k}" = "{replace_variable(v)}"'
                           for (k, v) in sorted(get_all_urls().items(), key=lambda x: x[0])])
    content = AUTO_GENERATED_COMMENT + 'export enum AppPath {\n' + app_urls + ',\n}\n'
    generate_file_if_updated('app paths', file_path, content, logger)


def generate_dajaxice_types():
    file_path = getattr(settings, 'REUSABLE_DAJAXICE_TS_INTERFACE_PATH', None)
    if not file_path or not dajaxice_functions:
        return
    from django_reusable.urls.utils import get_all_urls
    logger = PrintLogger("[django_reusable] generate_dajaxice_types")
    get_all_urls()  # needed for dajaxice to populate its registry
    modules = dict()
    for module, submodule in dajaxice_functions.modules.submodules.items():
        if not submodule.functions.items():
            continue
        modules[module] = [(name, [x for x in inspect.getfullargspec(func.function).args if x != 'request'])
                           for name, func in submodule.functions.items()]

    def get_args_type(args):
        if not args:
            return ''
        return '<{ ' + '; '.join([f"{arg}: any" for arg in args]) + ' }>'

    def get_functions(v):
        return ('{\n' + ';\n'.join([f'{spaces(8)}{name}: DajaxiceFn{get_args_type(args)}' for (name, args) in v])
                + f';\n{spaces(4)}' + '}')

    _modules = ';\n'.join([f'{spaces(4)}{k}?: {get_functions(v)}' for (k, v) in
                           sorted(modules.items(), key=lambda x: x[0])]) + ';'
    content = AUTO_GENERATED_COMMENT + DAJAXICE_STATIC_TYPES + 'export type DajaxiceModules = {\n' + _modules + '\n};\n'
    generate_file_if_updated('dajaxice types', file_path, content, logger)



from django.contrib import admin


class TextInputFilter(admin.SimpleListFilter):
    template = 'django_reusable/filters/input-filter.html'

    def lookups(self, request, model_admin):
        # Dummy, required to show the filter.
        return (('Dummy', 'Dummy'),)

    def queryset(self, request, queryset):
        return queryset


class SearchInFilter(admin.SimpleListFilter):
    title = 'Search In'
    parameter_name = 'search'
    lookup_choices = []

    def lookups(self, request, model_admin):
        return self.lookup_choices

    def queryset(self, request, queryset):
        return queryset



from datetime import datetime
from functools import wraps

from django.conf import settings
from django.core.cache import cache

from .utils import imap
from ..logging.loggers import PrintLogger


def get_decorator_key(func, func_name, by_args, enabled, *args, **kwargs):
    args_flattened = (', args:' + ','.join(imap(str, args)) + ','.join([f'{k}:{v}' for (k, v) in kwargs.items()])
                      if by_args and enabled else '')
    return f'dr:fn:{func_name or func.__name__}{args_flattened}'


def parametrized(decorator):
    """
    Decorator to wrap other decorators to preserve the arguments intellisense in IDEs.
    """

    def layer(*args, **kwargs):
        def repl(f):
            return decorator(f, *args, **kwargs)

        return repl

    return layer


@parametrized
def cache_data(func, timeout: int, by_args=False, enabled=True, custom_cache=None):
    """
    Decorator to cache the results of the method and return cached value if it's there. Useful for expensive calc.

    :param func: function to decorate
    :param timeout: Cache timeout.
    :param by_args: Caches by arguments if *True*.
    :param enabled: To enable or disable it.
    :param custom_cache: To provide a custom instance of cache instead of default django one
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not enabled:
            return func(*args, **kwargs)
        key = get_decorator_key(func, None, by_args, enabled, *args, **kwargs)
        _cache = custom_cache or cache
        return _cache.get_or_set(key, lambda: func(*args, **kwargs), timeout)

    return wrapper


@parametrized
def log_exec_time(func, func_name=None, enabled=getattr(settings, 'DR_LOG_EXEC_TIME', False), log_args=False):
    """
    Decorator to log calculate the execution time of the method.

    :param func: function to decorate
    :param func_name: custom function name to log
    :param enabled: To enable or disable it. Can be enabled globally by setting DR_LOG_EXEC_TIME=True in django settings.
    :param log_args: Log arguments passed if *True*.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        key = get_decorator_key(func, func_name, log_args, enabled, *args, **kwargs)
        logger = PrintLogger(f"log_exec_time ({key})", enabled=enabled)
        start = datetime.now()
        resp = func(*args, **kwargs)
        logger.info(f'execution time={(datetime.now() - start).seconds}s')
        return resp

    return wrapper
